@@ -33,10 +33,14 @@
 //! use skytable::{actions::Actions, Connection};
 //! let mut con = Connection::new("127.0.0.1", 2003).unwrap();
 //! con.set("x", "100").unwrap();
-//! assert_eq!(con.get("x").unwrap(), "100".to_owned());
-//!
+//! let ret: String = con.get("x").unwrap();
+//! assert_eq!(ret, "100");
 //! ```
+//!
 
+use crate::error::{errorstring, SkyhashError};
+use crate::types::Array;
+use crate::types::FromSkyhashBytes;
 use crate::types::SnapshotResult;
 use crate::Element;
 use crate::GetIterator;
@@ -44,76 +48,40 @@ use crate::IntoSkyhashAction;
 use crate::IntoSkyhashBytes;
 use crate::Query;
 use crate::RespCode;
-use crate::Response;
+use crate::SkyRawResult;
+use crate::SkyResult;
 
 cfg_async!(
     use core::{future::Future, pin::Pin};
 );
-use std::io::ErrorKind;
-
-/// The error string returned when the snapshot engine is busy
-pub const ERR_SNAPSHOT_BUSY: &str = "err-snapshot-busy";
-/// The error string returned when periodic snapshots are busy
-pub const ERR_SNAPSHOT_DISABLED: &str = "err-snapshot-disabled";
-
-/// Errors while running actions
-#[derive(Debug)]
-pub enum ActionError {
-    /// The server sent data but we failed to parse it
-    ParseError,
-    /// The server sent an unexpected data type for this action
-    UnexpectedDataType,
-    /// The server sent an unknown data type that we cannot parse
-    UnknownDataType,
-    /// The server sent an invalid response
-    InvalidResponse,
-    /// An I/O error occurred while running this action
-    IoError(ErrorKind),
-    /// The server returned a response code **other than the one that should have been returned
-    /// for this action** (if any)
-    Code(RespCode),
-}
 
 cfg_async!(
     /// A special result that is returned when running actions (async)
     pub type AsyncResult<'s, T> = Pin<Box<dyn Future<Output = T> + Send + Sync + 's>>;
     #[doc(hidden)]
+    /// A raw async connection to the database server
     pub trait AsyncSocket: Send + Sync {
-        fn run(&mut self, q: Query) -> AsyncResult<std::io::Result<Response>>;
+        /// Run the query
+        fn run(&mut self, q: Query) -> AsyncResult<SkyResult>;
     }
     impl<T> AsyncActions for T where T: AsyncSocket {}
 );
 
-/// A special result that is returned when running actions
-pub type ActionResult<T> = Result<T, ActionError>;
-
 cfg_sync!(
     #[doc(hidden)]
+    /// A raw synchronous connection to the database server
     pub trait SyncSocket {
-        fn run(&mut self, q: Query) -> std::io::Result<Response>;
+        /// Run the query
+        fn run(&mut self, q: Query) -> SkyResult;
     }
     impl<T> Actions for T where T: SyncSocket {}
 );
 
-macro_rules! gen_match {
-    ($ret:expr, $($($mtch:pat)+ $(if $exp:expr)*, $expect:expr),*) => {
-        match $ret {
-            $($(Ok($mtch))|* $(if $exp:expr)* => Ok($expect),)*
-            Ok(Response::InvalidResponse) => Err(ActionError::InvalidResponse),
-            Ok(Response::ParseError) => Err(ActionError::ParseError),
-            Ok(Response::UnsupportedDataType) => Err(ActionError::UnknownDataType),
-            Ok(Response::Item(Element::RespCode(code))) => Err(ActionError::Code(code)),
-            Ok(Response::Item(_)) => Err(ActionError::UnexpectedDataType),
-            Err(e) => Err(ActionError::IoError(e.kind())),
-        }
-    };
-}
-
 macro_rules! implement_actions {
     (
         $(
             $(#[$attr:meta])+
-            fn $name:ident$(<$($tyargs:ident : $ty:ident $(+$tye:lifetime)*),*>)?(
+            fn $name:ident$(<$($tyargs:ident $(: $ty:ident $(+$tye:lifetime)?)?),*>)?(
                 $($argname:ident: $argty:ty),*) -> $ret:ty {
                     $($block:block)?
                     $($($mtch:pat)|+ => $expect:expr),+
@@ -123,11 +91,22 @@ macro_rules! implement_actions {
         #[cfg(feature = "sync")]
         #[cfg_attr(docsrs, doc(cfg(feature = "sync")))]
         /// Actions that can be run on a [`SyncSocket`] connection
+        ///
+        /// ## Example
+        /// ```no_run
+        /// use skytable::actions::Actions;
+        /// use skytable::sync::Connection;
+        ///
+        /// let mut con = Connection::new("127.0.0.1", 2003).unwrap();
+        /// con.set("x", "100").unwrap();
+        /// let x: String = con.get("x").unwrap();
+        /// assert_eq!(x, "100");
+        /// ```
         pub trait Actions: SyncSocket {
             $(
                 $(#[$attr])*
                 #[inline]
-                fn $name<'s, $($($tyargs: $ty $(+$tye)*, )*)?>(&'s mut self $(, $argname: $argty)*) -> ActionResult<$ret> {
+                fn $name<'s, $($($tyargs$(: $ty $(+$tye)*,)* )*)?>(&'s mut self $(, $argname: $argty)*) -> SkyRawResult<$ret> {
                     gen_match!(self.run($($block)?), $($($mtch)+, $expect),*)
                 }
             )*
@@ -139,7 +118,7 @@ macro_rules! implement_actions {
             $(
                 $(#[$attr])*
                 #[inline]
-                fn $name<'s, $($($tyargs: $ty $(+$tye)*, )*)?>(&'s mut self $(, $argname: $argty)*) -> AsyncResult<ActionResult<$ret>> {
+                fn $name<'s, $($($tyargs$(: $ty $(+$tye)*,)* )*)?>(&'s mut self $(, $argname: $argty)*) -> AsyncResult<SkyRawResult<$ret>> {
                     Box::pin(async move {gen_match!(self.run($($block)?).await, $($($mtch)+, $expect),*)})
                 }
             )*
@@ -149,53 +128,84 @@ macro_rules! implement_actions {
 
 implement_actions!(
     /// Get the number of keys present in the database
-    fn dbsize() -> usize {
+    fn dbsize() -> u64 {
         { Query::from("dbsize") }
-        Response::Item(Element::UnsignedInt(int)) => int as usize
+        Element::UnsignedInt(int) => int
     }
     /// Deletes a single or a number of keys
     ///
+    /// This is equivalent to:
+    /// ```text
+    /// DEL <k1> <k2> <k3> ...
+    /// ```
+    ///
     /// This will return the number of keys that were deleted
-    fn del(key: impl IntoSkyhashAction + 's) -> usize {
+    ///
+    fn del(key: impl IntoSkyhashAction + 's) -> u64 {
         { Query::from("del").arg(key) }
-        Response::Item(Element::UnsignedInt(int)) => int as usize
+        Element::UnsignedInt(int) => int as u64
     }
     /// Checks if a key (or keys) exist(s)
     ///
+    /// This is equivalent to:
+    /// ```text
+    /// EXISTS <k1> <k2> <k3> ...
+    /// ```
+    ///
     /// This will return the number of keys that do exist
-    fn exists(key: impl IntoSkyhashAction + 's) -> usize {
+    ///
+    fn exists(key: impl IntoSkyhashAction + 's) -> u64 {
         { Query::from("exists").arg(key) }
-        Response::Item(Element::UnsignedInt(int)) => int as usize
+        Element::UnsignedInt(int) => int as u64
     }
     /// Removes all the keys present in the database
     fn flushdb() -> () {
         { Query::from("flushdb") }
-        Response::Item(Element::RespCode(RespCode::Okay)) => {}
+        Element::RespCode(RespCode::Okay) => {}
     }
     /// Get the value of a key
-    fn get(key: impl IntoSkyhashBytes + 's) -> String {
+    ///
+    /// This is equivalent to:
+    /// ```text
+    /// GET <key>
+    /// ```
+    fn get<T: FromSkyhashBytes>(key: impl IntoSkyhashBytes + 's) -> T {
         { Query::from("get").arg(key)}
-        Response::Item(Element::String(st)) => st
+        x @ Element::String(_) | x @ Element::Binstr(_) => T::from_element(x)?
     }
     /// Get the length of a key
-    fn keylen(key: impl IntoSkyhashBytes + 's) -> usize {
+    ///
+    /// This is equivalent to:
+    /// ```text
+    /// KEYLEN <key>
+    /// ```
+    fn keylen(key: impl IntoSkyhashBytes + 's) -> u64 {
         { Query::from("keylen").arg(key)}
-        Response::Item(Element::UnsignedInt(int)) => int as usize
+        Element::UnsignedInt(int) => int as u64
     }
     /// Returns a vector of keys
     ///
+    /// This is equivalent to:
+    /// ```text
+    /// LSKEYS <count>
+    /// ```
+    ///
     /// Do note that the order might be completely meaningless
-    fn lskeys(count: usize) -> Vec<String> {
-        { Query::from("lskeys").arg(count)}
-        Response::Item(Element::FlatArray(arr)) => arr
+    fn lskeys<T: FromSkyhashBytes>(count: u64) -> T {
+        { Query::from("lskeys").arg(count.to_string())}
+        x @ Element::Array(Array::Bin(_)) | x @ Element::Array(Array::Str(_)) => T::from_element(x)?
     }
     /// Get multiple keys
     ///
-    /// This returns a vector of [`Element`]s which either contain the values
-    /// as strings or contains `Not Found (Code: 1)` response codes
-    fn mget(keys: impl IntoSkyhashAction+ 's) -> Vec<Element> {
+    /// This returns a [`SimpleArray`] of values for the provided keys
+    ///
+    /// This is equivalent to:
+    /// ```text
+    /// MGET <k1> <k2> ...
+    /// ```
+    fn mget<T: FromSkyhashBytes>(keys: impl IntoSkyhashAction+ 's) -> T {
         { Query::from("mget").arg(keys)}
-        Response::Item(Element::Array(array)) => array
+        x @ Element::Array(Array::Bin(_)) | x @ Element::Array(Array::Str(_)) => T::from_element(x)?
     }
     /// Creates a snapshot
     ///
@@ -204,68 +214,119 @@ implement_actions!(
     /// which is normal behavior and _not an inherent error_
     fn mksnap() -> SnapshotResult {
        { Query::from("mksnap")}
-       Response::Item(Element::RespCode(RespCode::Okay)) => SnapshotResult::Okay,
-       Response::Item(Element::RespCode(RespCode::ErrorString(er))) => {
+       Element::RespCode(RespCode::Okay) => SnapshotResult::Okay,
+       Element::RespCode(RespCode::ErrorString(er)) => {
            match er.as_str() {
-               ERR_SNAPSHOT_BUSY => SnapshotResult::Busy,
-               ERR_SNAPSHOT_DISABLED => SnapshotResult::Disabled,
-               _ => return Err(ActionError::InvalidResponse)
+               errorstring::ERR_SNAPSHOT_BUSY => SnapshotResult::Busy,
+               errorstring::ERR_SNAPSHOT_DISABLED => SnapshotResult::Disabled,
+               _ => return Err(SkyhashError::InvalidResponse.into())
            }
        }
     }
+
     /// Sets the value of multiple keys and values and returns the number of keys that were set
     ///
+    /// This is equivalent to:
+    /// ```text
+    /// MSET <k1> <v1> <k2> <v2> ...
+    /// ```
+    /// with the only difference that you have to pass in the keys and values as separate
+    /// objects
+    ///
     /// ## Panics
     /// This method will panic if the number of keys and values are not equal
     fn mset<T: IntoSkyhashBytes + 's , U: IntoSkyhashBytes + 's>
     (
         keys: impl GetIterator<T> + 's,
         values: impl GetIterator<U> + 's
-    ) -> usize {
+    ) -> u64 {
         {
             assert!(keys.incr_len_by() == values.incr_len_by(), "The number of keys and values for mset must be equal");
             Query::from("mset")._push_alt_iter(keys, values)
         }
-        Response::Item(Element::UnsignedInt(int)) => int as usize
+        Element::UnsignedInt(int) => int as u64
     }
     /// Updates the value of multiple keys and values and returns the number of keys that were updated
     ///
+    /// This is equivalent to:
+    /// ```text
+    /// MUPDATE <k1> <v1> <k2> <v2> ...
+    /// ```
+    /// with the only difference that you have to pass in the keys and values as separate
+    /// objects
+    ///
     /// ## Panics
     /// This method will panic if the number of keys and values are not equal
     fn mupdate<T: IntoSkyhashBytes + 's , U: IntoSkyhashBytes + 's>
     (
         keys: impl GetIterator<T> + 's,
         values: impl GetIterator<U> + 's
-    ) -> usize {
+    ) -> u64 {
         {
             assert!(keys.incr_len_by() == values.incr_len_by(), "The number of keys and values for mupdate must be equal");
             Query::from("mset")._push_alt_iter(keys, values)
         }
-        Response::Item(Element::UnsignedInt(int)) => int as usize
+        Element::UnsignedInt(int) => int as u64
     }
     /// Consumes a key if it exists
     ///
-    /// This should return either the corresponding values of the provided keys or `Not Found`
-    /// error codes
-    fn pop(keys: impl IntoSkyhashAction + 's) -> Vec<Element> {
+    /// This will return the corresponding values of the provided key
+    /// depending on the type for that table
+    ///
+    /// This is equivalent to:
+    /// ```text
+    /// POP <key>
+    /// ```
+    fn pop<T: FromSkyhashBytes>(keys: impl IntoSkyhashBytes + 's) -> T {
         { Query::from("POP").arg(keys) }
-        Response::Item(Element::Array(arr)) => arr
+        x @ Element::String(_) | x @ Element::Binstr(_) => T::from_element(x)?
+    }
+    /// Consumes the provided keys if they exist
+    ///
+    /// This is equivalent to:
+    /// ```text
+    /// MPOP <k1> <k2> <k3>
+    /// ```
+    fn mpop<T: FromSkyhashBytes>(keys: impl IntoSkyhashAction + 's) -> T {
+        { Query::from("mpop").arg(keys)}
+        x @ Element::Array(Array::Bin(_)) | x @ Element::Array(Array::Str(_)) => T::from_element(x)?
     }
     /// Deletes all the provided keys if they exist or doesn't do anything at all. This method
     /// will return true if all the provided keys were deleted, else it will return false
+    ///
+    /// This is equivalent to:
+    /// ```text
+    /// SDEL <k1> <v1> <k2> <v2>
+    /// ```
+    /// with the only difference that you have to pass in the keys and values as separate
+    /// objects
     fn sdel(keys: impl IntoSkyhashAction + 's) -> bool {
         { Query::from("sdel").arg(keys) }
-        Response::Item(Element::RespCode(RespCode::Okay)) => true,
-        Response::Item(Element::RespCode(RespCode::NotFound)) => false
+        Element::RespCode(RespCode::Okay) => true,
+        Element::RespCode(RespCode::NotFound) => false
     }
     /// Set the value of a key
-    fn set(key: impl IntoSkyhashBytes + 's, value: impl IntoSkyhashBytes + 's) -> () {
+    ///
+    /// This is equivalent to:
+    /// ```text
+    /// SET <k> <v>
+    /// ```
+    ///
+    fn set(key: impl IntoSkyhashBytes + 's, value: impl IntoSkyhashBytes + 's) -> bool {
         { Query::from("set").arg(key).arg(value) }
-        Response::Item(Element::RespCode(RespCode::Okay)) => {}
+        Element::RespCode(RespCode::Okay) => true,
+        Element::RespCode(RespCode::OverwriteError) => false
     }
     /// Sets the value of all the provided keys or does nothing. This method will return true if all the keys
     /// were set or will return false if none were set
     ///
+    /// This is equivalent to:
+    /// ```text
+    /// SSET <k1> <v1> <k2> <v2> ...
+    /// ```
+    /// with the only difference that you have to pass in the keys and values as separate
+    /// objects
+    ///
     /// ## Panics
     /// This method will panic if the number of keys and values are not equal
     fn sset<T: IntoSkyhashBytes + 's , U: IntoSkyhashBytes + 's>
@@ -280,12 +341,18 @@ implement_actions!(
             );
             Query::from("sset")._push_alt_iter(keys, values)
         }
-        Response::Item(Element::RespCode(RespCode::Okay)) => true,
-        Response::Item(Element::RespCode(RespCode::OverwriteError)) => false
+        Element::RespCode(RespCode::Okay) => true,
+        Element::RespCode(RespCode::OverwriteError) => false
     }
     /// Updates the value of all the provided keys or does nothing. This method will return true if all the keys
-    /// were updated or will return false if none were updated
+    /// were updated or will return false if none were updated.
     ///
+    /// This is equivalent to:
+    /// ```text
+    /// SUPDATE <key1> <value1> <key2> <value2> ...
+    /// ```
+    /// with the only difference that you have to pass in the keys and values as separate
+    /// objects
     /// ## Panics
     /// This method will panic if the number of keys and values are not equal
     fn supdate<T: IntoSkyhashBytes + 's , U: IntoSkyhashBytes + 's>
@@ -300,23 +367,35 @@ implement_actions!(
             );
             Query::from("supdate")._push_alt_iter(keys, values)
         }
-        Response::Item(Element::RespCode(RespCode::Okay)) => true,
-        Response::Item(Element::RespCode(RespCode::NotFound)) => false
+        Element::RespCode(RespCode::Okay) => true,
+        Element::RespCode(RespCode::NotFound) => false
     }
     /// Update the value of a key
+    ///
+    /// This is equivalent to:
+    /// ```text
+    /// UPDATE <key> <value>
+    /// ```
     fn update(key: impl IntoSkyhashBytes + 's, value: impl IntoSkyhashBytes + 's) -> () {
         { Query::from("update").arg(key).arg(value) }
-        Response::Item(Element::RespCode(RespCode::Okay)) => {}
+        Element::RespCode(RespCode::Okay) => {}
     }
     /// Updates or sets all the provided keys and returns the number of keys that were set
     ///
+    /// This is equivalent to:
+    /// ```text
+    /// USET <key1> <value1> <key2> <value2> ...
+    /// ```
+    /// with the only difference that you have to pass in the keys and values as separate
+    /// objects
+    ///
     /// ## Panics
     /// This method will panic if the number of keys is not equal to the number of values
     fn uset<T: IntoSkyhashBytes + 's , U: IntoSkyhashBytes + 's>
     (
         keys: impl GetIterator<T> + 's,
         values: impl GetIterator<U> + 's
-    ) -> usize {
+    ) -> u64 {
         {
             assert!(
                 keys.incr_len_by() == values.incr_len_by(),
@@ -324,6 +403,6 @@ implement_actions!(
             );
             Query::from("uset")._push_alt_iter(keys, values)
         }
-        Response::Item(Element::UnsignedInt(int)) => int as usize
+        Element::UnsignedInt(int) => int as u64
     }
 );
