@@ -27,7 +27,11 @@
 //! by Sayan Nandan and this is the first client implementation of the protocol
 //!
 
+use crate::types::Array;
+use crate::types::FlatElement;
+use crate::types::FromSkyhashBytes;
 use crate::RespCode;
+use crate::SkyRawResult;
 use std::hint::unreachable_unchecked;
 
 #[derive(Debug)]
@@ -63,19 +67,27 @@ pub(super) struct Parser<'a> {
 ///
 /// This enum represents the data types supported by the Skyhash Protocol
 pub enum Element {
-    /// Arrays can be nested! Their `<tsymbol>` is `&`
-    Array(Vec<Element>),
-    /// A String value; `<tsymbol>` is `+`
+    /// Array types
+    Array(Array),
+    /// An unicode string value; `<tsymbol>` is `+`
     String(String),
+    /// A binary string (`?`)
+    Binstr(Vec<u8>),
     /// An unsigned integer value; `<tsymbol>` is `:`
     UnsignedInt(u64),
-    /// A non-recursive String array; tsymbol: `_`
-    FlatArray(Vec<String>),
     /// A response code
     RespCode(RespCode),
 }
 
+impl Element {
+    /// Try to convert an element to a type that implements [`FromSkyhashBytes`]
+    pub fn try_element_into<T: FromSkyhashBytes>(self) -> SkyRawResult<T> {
+        T::from_element(self)
+    }
+}
+
 #[derive(Debug, PartialEq)]
+#[non_exhaustive]
 /// # Parser Errors
 ///
 /// Several errors can arise during parsing and this enum accounts for them
@@ -92,7 +104,7 @@ pub enum ParseError {
     /// A data type was given but the parser failed to serialize it into this type
     ///
     /// This can happen not just for elements but can also happen for their sizes ([`Self::parse_into_u64`])
-    DataTypeParseError,
+    DataTypeError,
     /// A data type that the client doesn't know was passed into the query
     ///
     /// This is a frequent problem that can arise between different server editions as more data types
@@ -188,7 +200,7 @@ impl<'a> Parser<'a> {
         for dig in byte_iter {
             if !dig.is_ascii_digit() {
                 // dig has to be an ASCII digit
-                return Err(ParseError::DataTypeParseError);
+                return Err(ParseError::DataTypeError);
             }
             // 48 is the ASCII code for 0, and 57 is the ascii code for 9
             // so if 0 is given, the subtraction should give 0; similarly
@@ -200,11 +212,11 @@ impl<'a> Parser<'a> {
             // The usize can overflow; check that case
             let product = match item_usize.checked_mul(10) {
                 Some(not_overflowed) => not_overflowed,
-                None => return Err(ParseError::DataTypeParseError),
+                None => return Err(ParseError::DataTypeError),
             };
             let sum = match product.checked_add(curdig) {
                 Some(not_overflowed) => not_overflowed,
-                None => return Err(ParseError::DataTypeParseError),
+                None => return Err(ParseError::DataTypeError),
             };
             item_usize = sum;
         }
@@ -220,7 +232,7 @@ impl<'a> Parser<'a> {
         for dig in byte_iter {
             if !dig.is_ascii_digit() {
                 // dig has to be an ASCII digit
-                return Err(ParseError::DataTypeParseError);
+                return Err(ParseError::DataTypeError);
             }
             // 48 is the ASCII code for 0, and 57 is the ascii code for 9
             // so if 0 is given, the subtraction should give 0; similarly
@@ -232,11 +244,11 @@ impl<'a> Parser<'a> {
             // Now the entire u64 can overflow, so let's attempt to check it
             let product = match item_u64.checked_mul(10) {
                 Some(not_overflowed) => not_overflowed,
-                None => return Err(ParseError::DataTypeParseError),
+                None => return Err(ParseError::DataTypeError),
             };
             let sum = match product.checked_add(curdig) {
                 Some(not_overflowed) => not_overflowed,
-                None => return Err(ParseError::DataTypeParseError),
+                None => return Err(ParseError::DataTypeError),
             };
             item_u64 = sum;
         }
@@ -279,20 +291,62 @@ impl<'a> Parser<'a> {
             Err(ParseError::NotEnough)
         }
     }
-    /// The cursor should have passed the `+` tsymbol
-    fn parse_next_string(&mut self) -> ParseResult<String> {
-        let our_string_chunk = self.__get_next_element()?;
-        let our_string = String::from_utf8_lossy(&our_string_chunk).to_string();
+    /// The cursor should have passed the `?` tsymbol
+    fn parse_next_binstr(&mut self) -> ParseResult<Vec<u8>> {
+        let our_string_chunk = self.__get_next_element()?.to_owned();
         if self.will_cursor_give_linefeed()? {
-            // there is a lf after the end of the string; great!
+            // there is a lf after the end of the binary string; great!
             // let's skip that now
             self.incr_cursor();
             // let's return our string
-            Ok(our_string)
+            Ok(our_string_chunk)
         } else {
             Err(ParseError::UnexpectedByte)
         }
     }
+    /// Parse the next null checked element
+    fn parse_next_chunk_nullck(&mut self) -> ParseResult<Option<&[u8]>> {
+        // we have the chunk
+        let (start, stop) = self.read_line();
+        if let Some(sizeline) = self.buffer.get(start..stop) {
+            let string_size = Self::parse_into_usize_nullck(sizeline)?;
+            if let Some(size) = string_size {
+                // so it isn't null
+                let our_chunk = self.read_until(size)?;
+                Ok(Some(our_chunk))
+            } else {
+                Ok(None)
+            }
+        } else {
+            Err(ParseError::NotEnough)
+        }
+    }
+    /// The cursor should have passed the `+` tsymbol
+    fn parse_next_string(&mut self) -> ParseResult<String> {
+        Ok(String::from_utf8_lossy(&self.parse_next_binstr()?).to_string())
+    }
+    fn parse_next_binstr_nullck(&mut self) -> ParseResult<Option<Vec<u8>>> {
+        let our_chunk = self.parse_next_chunk_nullck()?;
+        if let Some(chunk) = our_chunk {
+            let our_chunk = chunk.to_owned();
+            if self.will_cursor_give_linefeed()? {
+                // there is a lf after the end of the binary string; great!
+                // let's skip that now
+                self.incr_cursor();
+                Ok(Some(our_chunk))
+            } else {
+                Err(ParseError::UnexpectedByte)
+            }
+        } else {
+            Ok(None)
+        }
+    }
+    fn parse_next_str_nullck(&mut self) -> ParseResult<Option<String>> {
+        match self.parse_next_binstr_nullck()? {
+            Some(chunk) => Ok(Some(String::from_utf8_lossy(&chunk).to_string())),
+            None => Ok(None),
+        }
+    }
     /// The cursor should have passed the `:` tsymbol
     fn parse_next_u64(&mut self) -> ParseResult<u64> {
         let our_u64_chunk = self.__get_next_element()?;
@@ -320,14 +374,31 @@ impl<'a> Parser<'a> {
     fn parse_next_element(&mut self) -> ParseResult<Element> {
         if let Some(tsymbol) = self.buffer.get(self.cursor) {
             // so we have a tsymbol; nice, let's match it
-            // but advance the cursor before doing that
+            // but advance the cursor before doing that (skip)
             self.incr_cursor();
             let ret = match *tsymbol {
+                b'?' => Element::Binstr(self.parse_next_binstr()?),
                 b'+' => Element::String(self.parse_next_string()?),
                 b':' => Element::UnsignedInt(self.parse_next_u64()?),
-                b'&' => Element::Array(self.parse_next_array()?),
+                b'&' => Element::Array(Array::Recursive(self.parse_next_array()?)),
                 b'!' => Element::RespCode(self.parse_next_respcode()?),
-                b'_' => Element::FlatArray(self.parse_next_flat_array()?),
+                b'@' => {
+                    // hmmm, a typed array; let's check the tsymbol
+                    if let Some(array_type) = self.buffer.get(self.cursor) {
+                        // got tsymbol, let's skip it
+                        self.incr_cursor();
+                        match array_type {
+                            b'+' => Element::Array(Array::Str(self.parse_next_typed_array_str()?)),
+                            b'?' => Element::Array(Array::Bin(self.parse_next_typed_array_bin()?)),
+                            _ => return Err(ParseError::UnknownDatatype),
+                        }
+                    } else {
+                        // if we couldn't fetch a tsymbol, there wasn't enough
+                        // data left
+                        return Err(ParseError::NotEnough);
+                    }
+                }
+                b'_' => Element::Array(Array::Flat(self.parse_next_flat_array()?)),
                 _ => return Err(ParseError::UnknownDatatype),
             };
             Ok(ret)
@@ -336,18 +407,60 @@ impl<'a> Parser<'a> {
             Err(ParseError::NotEnough)
         }
     }
-    /// The cursor should have passed the tsymbol
-    fn parse_next_flat_array(&mut self) -> ParseResult<Vec<String>> {
+    /// Parse the next null checked usize
+    fn parse_into_usize_nullck(inp: &[u8]) -> ParseResult<Option<usize>> {
+        if inp == [0] {
+            Ok(None)
+        } else {
+            Ok(Some(Self::parse_into_usize(inp)?))
+        }
+    }
+    /// The cursor should have passed the `@+` chars
+    fn parse_next_typed_array_str(&mut self) -> ParseResult<Vec<Option<String>>> {
         let (start, stop) = self.read_line();
         if let Some(our_size_chunk) = self.buffer.get(start..stop) {
-            let array_size = Self::parse_into_usize(&our_size_chunk)?;
+            // so we have a size chunk; let's get the size
+            let array_size = Self::parse_into_usize(our_size_chunk)?;
+            let mut array = Vec::with_capacity(array_size);
+            for _ in 0..array_size {
+                // no tsymbol, just elements and their sizes
+                array.push(self.parse_next_str_nullck()?);
+            }
+            Ok(array)
+        } else {
+            Err(ParseError::NotEnough)
+        }
+    }
+    /// The cursor should have passed the `@?` chars
+    fn parse_next_typed_array_bin(&mut self) -> ParseResult<Vec<Option<Vec<u8>>>> {
+        let (start, stop) = self.read_line();
+        if let Some(our_size_chunk) = self.buffer.get(start..stop) {
+            // got size chunk, let's get the size
+            let array_size = Self::parse_into_usize(our_size_chunk)?;
+            let mut array = Vec::with_capacity(array_size);
+            for _ in 0..array_size {
+                array.push(self.parse_next_binstr_nullck()?);
+            }
+            Ok(array)
+        } else {
+            Err(ParseError::NotEnough)
+        }
+    }
+    /// The cursor should have passed the tsymbol
+    fn parse_next_flat_array(&mut self) -> ParseResult<Vec<FlatElement>> {
+        let (start, stop) = self.read_line();
+        if let Some(our_size_chunk) = self.buffer.get(start..stop) {
+            let array_size = Self::parse_into_usize(our_size_chunk)?;
             let mut array = Vec::with_capacity(array_size);
             for _ in 0..array_size {
                 if let Some(tsymbol) = self.buffer.get(self.cursor) {
                     // good, there is a tsymbol; move the cursor ahead
                     self.incr_cursor();
                     let ret = match *tsymbol {
-                        b'+' => self.parse_next_string()?,
+                        b'+' => FlatElement::String(self.parse_next_string()?),
+                        b'?' => FlatElement::Binstr(self.parse_next_binstr()?),
+                        b'!' => FlatElement::RespCode(self.parse_next_respcode()?),
+                        b':' => FlatElement::UnsignedInt(self.parse_next_u64()?),
                         _ => return Err(ParseError::UnknownDatatype),
                     };
                     array.push(ret);
@@ -421,3 +534,63 @@ impl<'a> Parser<'a> {
         }
     }
 }
+
+#[test]
+fn test_typed_str_array() {
+    let typed_array_packet = "*1\n@+3\n3\nthe\n3\ncat\n6\nmeowed\n".as_bytes();
+    let (parsed, forward) = Parser::new(typed_array_packet).parse().unwrap();
+    assert_eq!(forward, typed_array_packet.len());
+    assert_eq!(
+        parsed,
+        RawResponse::SimpleQuery(Element::Array(Array::Str(vec![
+            Some("the".to_owned()),
+            Some("cat".to_owned()),
+            Some("meowed".to_owned())
+        ])))
+    );
+}
+
+#[test]
+fn test_typed_bin_array() {
+    let typed_array_packet = "*1\n@?3\n3\nthe\n3\ncat\n6\nmeowed\n".as_bytes();
+    let (parsed, forward) = Parser::new(typed_array_packet).parse().unwrap();
+    assert_eq!(forward, typed_array_packet.len());
+    assert_eq!(
+        parsed,
+        RawResponse::SimpleQuery(Element::Array(Array::Bin(vec![
+            Some(Vec::from("the")),
+            Some(Vec::from("cat")),
+            Some(Vec::from("meowed"))
+        ])))
+    );
+}
+
+#[test]
+fn test_typed_bin_array_null() {
+    let typed_array_packet = "*1\n@?3\n3\nthe\n3\ncat\n\0\n".as_bytes();
+    let (parsed, forward) = Parser::new(typed_array_packet).parse().unwrap();
+    assert_eq!(forward, typed_array_packet.len());
+    assert_eq!(
+        parsed,
+        RawResponse::SimpleQuery(Element::Array(Array::Bin(vec![
+            Some(Vec::from("the")),
+            Some(Vec::from("cat")),
+            None
+        ])))
+    );
+}
+
+#[test]
+fn test_typed_str_array_null() {
+    let typed_array_packet = "*1\n@+3\n3\nthe\n3\ncat\n\0\n".as_bytes();
+    let (parsed, forward) = Parser::new(typed_array_packet).parse().unwrap();
+    assert_eq!(forward, typed_array_packet.len());
+    assert_eq!(
+        parsed,
+        RawResponse::SimpleQuery(Element::Array(Array::Str(vec![
+            Some("the".to_owned()),
+            Some("cat".to_owned()),
+            None
+        ])))
+    );
+}
