@@ -38,7 +38,7 @@
 //!
 //! First add this to your `Cargo.toml` file:
 //! ```toml
-//! skytable = "0.5.0"
+//! skytable = "0.6.0"
 //! ```
 //! Now open up your `src/main.rs` file and establish a connection to the server while also adding some
 //! imports:
@@ -61,6 +61,36 @@
 //!     Ok(())
 //! }
 //! ```
+//!
+//! ## Running actions
+//!
+//! As noted [below](#binary-data), the default table is a key/value table with a binary key
+//! type and a binary value type. Let's go ahead and run some actions (we're assuming you're
+//! using the sync API; for async, simply change the import to `use skytable::actions::AsyncActions`).
+//!
+//! ### `SET`ting a key
+//!
+//! ```no_run
+//! use skytable::actions::Actions;
+//! use skytable::sync::Connection;
+//!
+//! let mut con = Connection::new("127.0.0.1", 2003).unwrap();
+//! con.set("hello", "world").unwrap();
+//! ```
+//!
+//! This will set the value of the key `hello` to `world` in the `default:default` entity.
+//!
+//! ### `GET`ting a key
+//!
+//! ```no_run
+//! use skytable::actions::Actions;
+//! use skytable::sync::Connection;
+//!
+//! let mut con = Connection::new("127.0.0.1", 2003).unwrap();
+//! let x: String = con.get("hello").unwrap();
+//! assert_eq!(x, "world");
+//! ```
+//!
 //! Way to go &mdash; you're all set! Now go ahead and run more advanced queries!
 //!
 //! ## Binary data
@@ -88,13 +118,14 @@
 //!
 //! Now that you know how you can run basic queries, check out the [`actions`] module documentation for learning
 //! to use actions and the [`types`] module documentation for implementing your own Skyhash serializable
-//! types. You can also find the [latest examples here](https://github.com/skytable/client-rust/tree/next/examples)
+//! types. Need to meddle with DDL queries like creating and dropping tables? Check out the [`ddl`] module.
+//! You can also find the [latest examples here](https://github.com/skytable/client-rust/tree/next/examples)
 //!
 //! ## Async API
 //!
 //! If you need to use an `async` API, just change your import to:
 //! ```toml
-//! skytable = { version = "0.5.0", features=["async"], default-features=false }
+//! skytable = { version = "0.6.0", features=["async"], default-features=false }
 //! ```
 //! You can now establish a connection by using `skytable::AsyncConnection::new()`, adding `.await`s wherever
 //! necessary. Do note that you'll the [Tokio runtime](https://tokio.rs).
@@ -104,7 +135,7 @@
 //! With this client driver, it is possible to use both sync and `async` APIs **at the same time**. To do
 //! this, simply change your import to:
 //! ```toml
-//! skytable = { version="0.5.0", features=["sync", "async"] }
+//! skytable = { version="0.6.0", features=["sync", "async"] }
 //! ```
 //!
 //! ## TLS
@@ -113,13 +144,13 @@
 //!
 //! ### Using TLS with sync interfaces
 //! ```toml
-//! skytable = { version="0.5.0", features=["sync","ssl"] }
+//! skytable = { version="0.6.0", features=["sync","ssl"] }
 //! ```
 //! You can now use the async [TlsConnection](`sync::TlsConnection`) object.
 //!
 //! ### Using TLS with async interfaces
 //! ```toml
-//! skytable = { version="0.5.0", features=["async","aio-ssl"], default-features=false }
+//! skytable = { version="0.6.0", features=["async","aio-ssl"], default-features=false }
 //! ```
 //! You can now use the async [TlsConnection](`aio::TlsConnection`) object.
 //!
@@ -142,13 +173,20 @@
 //!
 
 #![cfg_attr(docsrs, feature(doc_cfg))]
+// macro mods
 #[macro_use]
 mod util;
+// endof macro mods
+// public mods
 pub mod actions;
-mod deserializer;
+pub mod ddl;
 pub mod error;
-mod respcode;
 pub mod types;
+// endof public mods
+// private mods
+mod deserializer;
+mod respcode;
+// endof private mods
 use crate::types::GetIterator;
 pub use deserializer::Element;
 pub use respcode::RespCode;
@@ -156,6 +194,13 @@ pub(crate) use std::io::Result as IoResult;
 use types::IntoSkyhashAction;
 use types::IntoSkyhashBytes;
 
+/// The default host address
+pub const DEFAULT_HOSTADDR: &str = "127.0.0.1";
+/// The default port
+pub const DEFAULT_PORT: u16 = 2003;
+/// The default entity
+pub const DEFAULT_ENTITY: &str = "default:default";
+
 cfg_async!(
     pub mod aio;
     pub use aio::Connection as AsyncConnection;
@@ -171,7 +216,7 @@ pub type SkyRawResult<T> = Result<T, self::error::Error>;
 /// A specialized error type for queries
 pub type SkyResult = SkyRawResult<Element>;
 
-#[derive(Debug)]
+#[derive(Debug, Clone)]
 /// A connection builder for easily building connections
 ///
 /// ## Example (sync)
@@ -179,8 +224,9 @@ pub type SkyResult = SkyRawResult<Element>;
 /// use skytable::ConnectionBuilder;
 /// let con =
 ///     ConnectionBuilder::new()
-///     .set_host("127.0.0.1")
+///     .set_host("127.0.0.1".to_string())
 ///     .set_port(2003)
+///     .set_entity("mykeyspace:mytable".to_string())
 ///     .get_connection()
 ///     .unwrap();
 /// ```
@@ -193,45 +239,55 @@ pub type SkyResult = SkyRawResult<Element>;
 ///         ConnectionBuilder::new()
 ///         .set_host("127.0.0.1")
 ///         .set_port(2003)
+///         .set_entity("mykeyspace:mytable".to_string())
 ///         .get_async_connection()
 ///         .await
 ///         .unwrap();
 /// }
 /// ```
-pub struct ConnectionBuilder<'a> {
+pub struct ConnectionBuilder {
     port: Option<u16>,
-    host: Option<&'a str>,
+    host: Option<String>,
+    entity: Option<String>,
 }
 
-impl<'a> Default for ConnectionBuilder<'a> {
+impl Default for ConnectionBuilder {
     fn default() -> Self {
         Self::new()
     }
 }
 
-impl<'a> ConnectionBuilder<'a> {
+impl ConnectionBuilder {
     /// Create an empty connection builder
     pub fn new() -> Self {
         Self {
             port: None,
             host: None,
+            entity: None,
         }
     }
-    /// Set the port
+    /// Set the port (defaults to `2003`)
     pub fn set_port(mut self, port: u16) -> Self {
         self.port = Some(port);
         self
     }
-    /// Set the host
-    pub fn set_host(mut self, host: &'a str) -> Self {
+    /// Set the host (defaults to `localhost`)
+    pub fn set_host(mut self, host: String) -> Self {
         self.host = Some(host);
         self
     }
+    /// Set the entity (defaults to `default:default`)
+    pub fn set_entity(mut self, entity: String) -> Self {
+        self.entity = Some(entity);
+        self
+    }
     cfg_sync! {
         /// Get a [sync connection](sync::Connection) to the database
         pub fn get_connection(&self) -> SkyRawResult<sync::Connection> {
-            let con =
-                sync::Connection::new(self.host.unwrap_or("127.0.0.1"), self.port.unwrap_or(2003))?;
+            use crate::ddl::Ddl;
+            let mut con =
+                sync::Connection::new(self.host.as_ref().unwrap_or(&DEFAULT_HOSTADDR.to_owned()), self.port.unwrap_or(2003))?;
+            con.switch(self.entity.as_ref().unwrap_or(&DEFAULT_ENTITY.to_owned()))?;
             Ok(con)
         }
         cfg_sync_ssl_any! {
@@ -240,11 +296,13 @@ impl<'a> ConnectionBuilder<'a> {
                 &self,
                 sslcert: String,
             ) -> SkyRawResult<sync::TlsConnection> {
-                let con = sync::TlsConnection::new(
-                    self.host.unwrap_or("127.0.0.1"),
+                use crate::ddl::Ddl;
+                let mut con = sync::TlsConnection::new(
+                    self.host.as_ref().unwrap_or(&DEFAULT_HOSTADDR.to_owned()),
                     self.port.unwrap_or(2003),
                     &sslcert,
                 )?;
+                con.switch(self.entity.as_ref().unwrap_or(&DEFAULT_ENTITY.to_owned()))?;
                 Ok(con)
             }
         }
@@ -252,8 +310,10 @@ impl<'a> ConnectionBuilder<'a> {
     cfg_async! {
         /// Get an [async connection](aio::Connection) to the database
         pub async fn get_async_connection(&self) -> SkyRawResult<aio::Connection> {
-            let con = aio::Connection::new(self.host.unwrap_or("127.0.0.1"), self.port.unwrap_or(2003))
+            use crate::ddl::AsyncDdl;
+            let mut con = aio::Connection::new(self.host.as_ref().unwrap_or(&DEFAULT_HOSTADDR.to_owned()), self.port.unwrap_or(2003))
                 .await?;
+            con.switch(self.entity.as_ref().unwrap_or(&DEFAULT_ENTITY.to_owned())).await?;
             Ok(con)
         }
         cfg_async_ssl_any! {
@@ -262,12 +322,14 @@ impl<'a> ConnectionBuilder<'a> {
                 &self,
                 sslcert: String,
             ) -> SkyRawResult<aio::TlsConnection> {
-                let con = aio::TlsConnection::new(
-                    self.host.unwrap_or("127.0.0.1"),
+                use crate::ddl::AsyncDdl;
+                let mut con = aio::TlsConnection::new(
+                    self.host.as_ref().unwrap_or(&DEFAULT_HOSTADDR.to_owned()),
                     self.port.unwrap_or(2003),
                     &sslcert,
                 )
                 .await?;
+                con.switch(self.entity.as_ref().unwrap_or(&DEFAULT_ENTITY.to_owned())).await?;
                 Ok(con)
             }
         }
@@ -385,6 +447,10 @@ impl Query {
     /// ## Panics
     /// This method will panic if the passed `arg` is empty
     pub fn push(&mut self, arg: impl IntoSkyhashAction) {
+        assert!(
+            arg.incr_len_by() != 0,
+            "An argument passed to a query cannot be empty!"
+        );
         arg.push_into_query(self);
     }
     pub(in crate) fn _push_alt_iter<T, U>(
@@ -397,8 +463,8 @@ impl Query {
         U: IntoSkyhashBytes,
     {
         v1.get_iter().zip(v2.get_iter()).for_each(|(a, b)| {
-            self._push_arg(a.to_bytes());
-            self._push_arg(b.to_bytes());
+            self._push_arg(a.as_bytes());
+            self._push_arg(b.as_bytes());
         });
         self
     }
